from typing import Any, Pattern, Union

import records.extras as extras
from records.fillers.validators import AssertValidation, GlobalValidationToken


class _Least:  # pragma: no cover
    """
    A simple object that is always less than any other object
    """

    def __lt__(self, other):
        return True

    def __le__(self, other):
        return True

    def __gt__(self, other):
        return False

    def __ge__(self, other):
        return self == other


class _Greatest:  # pragma: no cover
    """
    A simple object that is always greater than any other object
    """

    def __lt__(self, other):
        return False

    def __le__(self, other):
        return self == other

    def __gt__(self, other):
        return True

    def __ge__(self, other):
        return True


least = _Least()
greatest = _Greatest()


class Clamp(GlobalValidationToken):
    """
    A validator class that constrains the value to be between two bounds, bringing it to the nearest bound if it
     falls outside.
    """

    def __init__(self, ge: Any = least, le: Any = greatest, **kwargs):
        """
        :param ge: the lower bound, defaults to no lower bound
        :param le: the upper bound, defaults to no upper bound
        """
        super().__init__(**kwargs)
        self.ge = ge
        self.le = le
        if self.ge > self.le:
            raise ValueError('the lower bound must not be greater then the upper bound')

    def inner(self, v):
        return min(max(v, self.ge), self.le)

    def __call__(self, *_):
        return self.inner


class Cyclic(GlobalValidationToken):
    """
   A validator class that constrains the value to be between two bounds, bringing it to the equivalent position as
    though the domain is cyclic. Useful for angles and time of day.
   """

    def __init__(self, minimum, maximum, **kwargs):
        """
        :param minimum: the inclusive lower bound
        :param maximum: the exclusive upper bound.
        """
        super().__init__(**kwargs)
        self.minimum = minimum
        self.maximum = maximum

    def inner(self, v):
        if self.minimum <= v < self.maximum:
            return v
        d = v - self.minimum
        d %= (self.maximum - self.minimum)
        return self.minimum + d

    def __call__(self, *_):
        return self.inner


class Within(AssertValidation):
    """
    An assertion validation that raises an error if the value falls outside of bounds.
    """

    def __init__(self, ge: Any = least, lt: Any = greatest, g_eq=True, l_eq=False, **kwargs):
        """
        :param ge: the lower bound, defaults to no lower bound
        :param le: the upper bound, defaults to no upper bound
        :param g_eq: whether the lower bound is inclusive, defaults to True.
        :param l_eq: whether the upper bound is inclusive, defaults to False.
        :param kwargs: forwarded to `AssertValidation`_
        """
        super().__init__(**kwargs)
        self.ge = ge
        self.lt = lt
        self.g_eq = g_eq
        self.l_eq = l_eq
        if ge > lt or ((not g_eq or not l_eq) and ge == lt):
            raise ValueError('the lower bound must not be greater then the upper bound')

    def assert_(self, v) -> bool:
        if self.g_eq:
            if not (self.ge <= v):
                return False
        else:
            if not (self.ge < v):
                return False

        if self.l_eq:
            if not (self.lt >= v):
                return False
        else:
            if not (self.lt > v):
                return False

        return True


class FullMatch(AssertValidation):
    """
    An assertion validation that raises an error if the value does not match a regex pattern.
    """

    def __init__(self, pattern: Union[Pattern, str, bytes], **kwargs):
        """
        :param pattern: either a compiled pattern or an uncompiled string or bytestring
        :param kwargs: forwarded to `AssertValidation`_
        .. note::
            ``pattern`` will be compiled in accordance to `records.extras.re`_
        """
        super().__init__(**kwargs)
        self.pattern = extras.re.compile(pattern)

    def assert_(self, v) -> bool:
        return bool(self.pattern.fullmatch(v))


class Truth(AssertValidation):
    """
    An assertion validation that raises an error if the value does not evaluate as True.
    """

    def assert_(self, v) -> bool:
        return bool(v)



from abc import ABC, abstractmethod
from ast import literal_eval
from functools import partial, wraps
from inspect import isabstract
from math import isclose
from numbers import Complex, Rational, Real, Integral
from typing import Any, Callable, Dict, Generic, Type, TypeVar, Union, Tuple, Optional

from records.fillers.builtin_fillers.recurse import GetFiller
from records.fillers.coercers import CoercionToken, GlobalCoercionToken
from records.fillers.filler import AnnotatedFiller, TypeCheckStyle, TypeMatch

T = TypeVar('T')


class Eval(GlobalCoercionToken):
    """
    A coercion token to evaluate a string input as a python expression using `eval`
    .. warning::
        evaluating arbitrary strings is always risky!
    """

    def __init__(self, *args, **kwargs):
        if args and args[0] is ...:
            self.add_bound = True
            args = args[1:]
        else:
            self.add_bound = False
        self.kwargs = kwargs
        self.kwargs.update(
            (k.__name__, k) for k in args
        )

    def __call__(self, origin, filler):
        ns = self.kwargs
        if self.add_bound:
            ns = {**self.kwargs, origin.__name__: origin}

        def ret(v):
            if not isinstance(v, str):
                raise TypeError
            try:
                ret = eval(v, {'__builtins__': ns})
            except Exception as e:
                raise TypeError from e

            return ret

        return ret


class LiteralEval(GlobalCoercionToken):
    """
    A coercion token to evaluate string inputs as literals continuously
    """

    def __call__(self, origin, filler):
        if origin not in {int, str, float, complex, set, tuple, dict, list, bool}:
            raise TypeError(f'cannot use literal evaluation coercer with non-literal type {origin}')

        def ret(v):
            while isinstance(v, str):
                try:
                    v = literal_eval(v)
                except ValueError:
                    raise TypeError
            return v

        return ret


class OriginDependant(GlobalCoercionToken, ABC):
    """
    A convenience coercion token superclass to pass coercion directly to a token method
    """

    def __init__(self, *args, **kwargs):
        """
        :param args: arguments accessible to the callback
        :param kwargs: keywords accessible to the callback
        """
        self.args = args
        self.kwargs = kwargs

    @staticmethod
    @abstractmethod
    def func_args(origin, v):
        """
        The coercion callback to apply on an argument, given an origin.
        :param origin: The target storage type.
        :param v: The input argument to the coercer.
        :return: The coerced value
        """
        pass

    def __call__(self, origin, filler):
        return partial(self.func_args, origin)


class LooseMixin(OriginDependant, ABC):
    """
    A superclass for "loose" coercion tokens that call the class constructor, forbidding binding to abstract classes
    """

    def __call__(self, origin, filler):
        if (not isinstance(origin, type)) or isabstract(origin):
            raise TypeError(f'cannot use loose coercer with non-concrete type {origin}')
        return super().__call__(origin, filler)


class Loose(LooseMixin):
    """
    A coercion token to call the class constructor with the input as an argument
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # If type_constraint is set, only arguments of that type would be accepted as inputs
        self._type_constraint: Union[None, type, Tuple[type, ...]] = None

    def func_args(self, origin, v):
        if self._type_constraint and not isinstance(v, self._type_constraint):
            raise TypeError(f'must be of type {self._type_constraint}')
        return origin(v, *self.args, **self.kwargs)

    @classmethod
    def constrain(cls, item: Union[type, Tuple[type, ...]]):
        """
        Can be used to constrain Loose to only accept specific inputs of specific types.
        :param item: the type or types to constrain inputs by.
        :return: a factory function to a constrained Loose token.
        """
        # here we check that the item can indeed by used for type checking
        # we need to thoroughly check the input type here since if there's an error during coercion,
        # we won't know about it
        isinstance(None, item)

        @wraps(cls)
        def factory(*args, **kwargs):
            ret = cls(*args, **kwargs)
            ret._type_constraint = item
            return ret

        return factory


class LooseUnpack(LooseMixin):
    """
    A coercion token to call the class constructor with the input as an unpacked iterable
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # If type_constraint is set, only arguments of that type would be accepted as inputs
        self._type_constraints: Union[None, Tuple[Union[type, Tuple[type, ...]], ...],
                                      Tuple[Union[type, Tuple[type, ...]], type(...)]] = None

    def func_args(self, origin, v):
        if self._type_constraints is not None:
            v = tuple(v)
            if len(self._type_constraints) == 2 and self._type_constraints[1] is ...:
                if any(not isinstance(i, self._type_constraints[0]) for i in v):
                    raise TypeError(f'all arguments must be instances of {self._type_constraints[0]}')
            else:
                if len(v) != len(self._type_constraints) \
                        or any(not isinstance(i, tc) for (i, tc) in zip(v, self._type_constraints)):
                    raise TypeError
        return origin(*v, *self.args, **self.kwargs)

    @classmethod
    def constrain(cls, *items: Union[type, Tuple[type, ...], type(...)]):
        """
        Can be used to constrain LooseUnpack to only accept specific inputs of specific types.
        :param items: the type or types to constrain inputs by.
        :return: a factory function to a constrained LooseUnpack token.
        """
        # here we check that the item can indeed by used for type checking
        if len(items) == 2 and items[1] is ...:
            isinstance(None, items[0])
        else:
            for i in items:
                isinstance(None, i)

        @wraps(cls)
        def factory(*args, **kwargs):
            ret = cls(*args, **kwargs)
            ret._type_constraints = items
            return ret

        return factory


class LooseUnpackMap(LooseMixin):
    """
    A coercion token to call the class constructor with the input as an unpacked mapping
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # If type_constraint is set, only arguments of that type would be accepted as inputs
        self._type_constraints: Optional[Dict[str, Union[type, Tuple[type, ...]]]] = None

    def func_args(self, origin, v):
        if self._type_constraints is not None:
            for k, a in v.items():
                tc = self._type_constraints.get(k)
                if tc is None or not isinstance(a, tc):
                    raise TypeError
        return origin(*self.args, **v, **self.kwargs)

    @classmethod
    def constrain(cls, **items: Union[type, Tuple[type, ...]]):
        """
        Can be used to constrain LooseUnpackMap to only accept specific inputs of specific types.
        :param items: the type or types to constrain inputs by.
        :return: a factory function to a constrained LooseUnpackMap token.
        """
        # here we check that the item can indeed by used for type checking
        for v in items.values():
            isinstance(None, v)

        @wraps(cls)
        def factory(*args, **kwargs):
            ret = cls(*args, **kwargs)
            ret._type_constraints = items
            return ret

        return factory


class SimpleFiller(AnnotatedFiller[T], Generic[T]):
    """
    A concrete filler class that uses instance checking to check types
    """

    def type_check(self, v):
        if type(v) == self.origin:
            return TypeMatch.exact
        return TypeMatch.inexact if isinstance(v, self.origin) else None

    def bind(self, owner_cls):
        super().bind(owner_cls)

        if self.type_checking_style == TypeCheckStyle.check_strict and isabstract(self.origin):
            raise TypeError(f'cannot create strict checker for abstract class {self.origin}')


class Whole(OriginDependant):
    """
    A coercion token to attempt to convert a whole Number to an Integer type
    """

    def func_args(self, origin, v):
        if isinstance(v, Rational):
            if v.denominator == 1:
                return origin(v.numerator, *self.args, **self.kwargs)
        elif isinstance(v, Real):
            mod = v % 1
            if isclose(mod, 0) or isclose(mod, 1):
                return origin(v, *self.args, **self.kwargs)
        elif isinstance(v, Complex):
            if v.imag == 0:
                return self.func_args(origin, v.real)
        raise TypeError


class ToBytes(OriginDependant):
    """
    A coercion token to convert a n Integer value to a bytestring
    """

    def func_args(self, origin, v):
        # In theory we aught to restrict this token to just ints, but if another Integer subclass wants to use this
        # coercer we aren't gonna stop them
        if not isinstance(v, Integral):
            raise type
        if not self.args and ('length' not in self.kwargs):
            signed = self.kwargs.get('signed', False)
            args = ((v.bit_length() + 7 + signed) // 8,)
        else:
            args = self.args
        return origin(v.to_bytes(*args, **self.kwargs))


class FromInteger(CoercionToken):
    """
    A coercion token to convert an integer to a boolean, failing if the value is not 0 or 1.
    """
    pass


class BoolFiller(SimpleFiller[bool]):
    """
    A specialized boolean filler to handle `FromInteger`
    """

    @staticmethod
    def _bool_from_int(v):
        if v == 0:
            return False
        if v == 1:
            return True
        raise TypeError

    def get_coercer(self, token):
        if isinstance(token, FromInteger):
            return self._bool_from_int
        return super().get_coercer(token)


class Falsish(OriginDependant):
    """
    A coercion token to construct an instance of the target type ignoring the input, only if the argument is falsish.
     Useful to create empty objects from None inputs.
    """

    def func_args(self, origin, v):
        if not v:
            return origin(*self.args, **self.kwargs)
        raise TypeError


class NoneFiller(SimpleFiller[None]):
    """
    A specialized filler for None
    """

    def type_check(self, v) -> bool:
        return (v is None) and TypeMatch.exact


class EllipsisFiller(SimpleFiller[type(...)]):
    """
    A specialized filler for Ellipsis
    """

    def type_check(self, v) -> bool:
        return (v is ...) and TypeMatch.exact


class CallableFiller(AnnotatedFiller[Callable]):
    """
    A specialized filler for callable objects
    """

    def type_check(self, v):
        return callable(v) and TypeMatch.exact


std_filler_map: Dict[Any, Type[AnnotatedFiller]] = {
    bool: BoolFiller,
    type(None): NoneFiller,
    object: SimpleFiller,
    type(...): EllipsisFiller
}

std_filler_checkers = []


@std_filler_checkers.append
def callable_checker(stored_type):
    if stored_type is callable:
        return CallableFiller


@std_filler_checkers.append
def none_checker(stored_type):
    if stored_type is None:
        return GetFiller(type(None))



from typing import Any, Callable, Dict, List, Type, Union

from records.fillers.builtin_fillers.recurse import GetFiller
from records.fillers.filler import AnnotatedFiller

builtin_filler_map: Dict[Any, Type[AnnotatedFiller]] = {}
"""
a mapping of origin types to builtin filler types. If an exact match is not found in the map,
 then the checkers below are called, if none of them match, the mapping is checked again for subclassing.
"""
builtin_filler_checkers: List[Callable[[Any], Union[None, GetFiller, Type[AnnotatedFiller]]]] = []
"""
A sequence of callbacks to check if a storage type can be fitted to a filler type. Each callback should return `None`
 if there is no match, a `Filler` subclass if there is a match, or an instance of `GetFiller` for redirections.
"""



from records.field import Factory
from records.fillers.builtin_fillers.std_fillers import (LiteralEval, Eval, FromInteger, Loose, LooseUnpack,
                                                         LooseUnpackMap, Falsish, Whole)
from records.fillers.builtin_validators import Clamp, Cyclic, FullMatch, Truth, Within
from records.fillers.coercers import CallCoercion, ClassMethodCoercion, ComposeCoercer, MapCoercion
from records.fillers.filler import TypeCheckStyle
from records.fillers.validators import AssertCallValidation, AssertValidation, CallValidation, ValidationToken
from records.record import RecordBase, parser, exclude_from_ordering
from records.select import SelectableFactory
from records.tags import Tag
from records.utils.typing_compatible import Annotated

check = TypeCheckStyle.check
check_strict = TypeCheckStyle.check_strict
hollow = TypeCheckStyle.hollow

__all__ = [
    'RecordBase', 'Annotated', 'Factory', 'parser', 'exclude_from_ordering',

    'TypeCheckStyle', 'check', 'check_strict', 'hollow',

    'CallCoercion', 'MapCoercion', 'ComposeCoercer', 'ClassMethodCoercion',
    'ValidationToken', 'AssertValidation', 'CallValidation', 'AssertCallValidation',
    'Clamp', 'Within', 'FullMatch', 'Truth', 'Cyclic',

    'Tag',
    'SelectableFactory',

    'Eval', 'LiteralEval', 'Loose', 'LooseUnpack', 'LooseUnpackMap', 'Whole', 'FromInteger',
    'Falsish'
]

# TODO:
#  * fill on assign? post_init_on assign? invariant?!
#  * more examples
#  * add argument to Loose to only accept inputs of certain types
#  * typevars?
#  * ways to add validators to sub-parsers
#  * allow multiple validators to succeed in Unions if they produce the same value
#  * todo fill trace-back (Union[Annotated[int, Negative], Annotated[bool, Loose]])



from abc import ABC, abstractmethod
from typing import Any, Callable, Generic, Mapping, Optional, Type, TypeVar, Union

from records.fillers.util import _as_instance


class CoercionToken:
    """
    A base class for all object that should be interpreted as coercers
    """
    pass


T = TypeVar('T')


class GlobalCoercionToken(CoercionToken, ABC):
    """
    A base class for all coercers that can act on multiple kinds of fillers
    """

    @abstractmethod
    def __call__(self, stored_type: Type[T], filler) -> Callable[[Any], T]:
        """
        :param stored_type: the type the callback should return
        :param filler: the filler the coercion callback will be called from
        :return: a callable to act as a coercion callback for the specified filler
        """
        pass


class CallCoercion(GlobalCoercionToken, Generic[T]):
    """
    A coercion token to call arbitrary functions
    """

    def __init__(self, func: Callable[..., T], *args, **kwargs):
        """
        :param func: The callable to use as the coercion callback.
        :param args: Optional positional arguments to pass to ``func``, after the validation argument.
        :param kwargs: Optional keyword arguments to pass to ``func``.
        .. note::
            calling ``CallCoercion`` with ``args`` or ``kwargs`` is akin to calling it with a `functools.partial`_ as
             ``func``.
            >>> CallCoercion(foo, a, b, c=d)
            >>> # is equivalent to
            >>> CallCoercion(lambda v: foo(v, a, b, c=d))
        """
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def inner(self, v):
        return self.func(v, *self.args, **self.kwargs)

    def __call__(self, *_):
        return self.inner


class MapCoercion(GlobalCoercionToken, Generic[T]):
    """
    A coercion token to map values by arbitrary, pre-defined mappings
    """

    def __init__(self, value_map: Optional[Mapping[Any, T]] = None,
                 factory_map: Optional[Mapping[Any, Callable[[], T]]] = None):
        """
        :param value_map: a mapping to map values to coerced values
        :param factory_map: a mapping to map values to factory callbacks that create coerced values
        """
        self.value_map = value_map or {}
        self.factory_map = factory_map or {}

    def inner(self, v):
        if v in self.value_map:
            return self.value_map[v]
        if v in self.factory_map:
            return self.factory_map[v]()
        raise TypeError

    def __call__(self, *_):
        return self.inner


class ClassMethodCoercion(GlobalCoercionToken, Generic[T]):
    """
    A coercion token to call a class method in the target class
    """

    def __init__(self, method: str, *args, **kwargs):
        """
        :param method: The name of the class method to call
        :param args: Optional positional arguments to pass to the class method, after the validation argument.
        :param kwargs: Optional keyword arguments to pass to class method.
        """
        self.method = method
        self.args = args
        self.kwargs = kwargs

    def __call__(self, cls, filler):
        func = getattr(cls, self.method)

        def ret(v):
            return func(v, *self.args, **self.kwargs)

        return ret


class ComposeCoercer(GlobalCoercionToken):
    """
    A coercion token to chain two coercion callbacks one after the other
    """

    def __init__(self, *inner_coercers: Union[Type[CoercionToken], CoercionToken]):
        """
        :param inner_coercers: an iterable of coercion tokens to apply upon the argument, in reverse order.
        .. note::
            the inner coercion callbacks are called in reverse order. So if a token ``A`` will result in callback ``a``,
            and token ``B`` will result in callback ``b``, then the token ``ComposeCoercer(A,B)`` will result in
            callback ``lambda v: a(b(v))``.
        """
        self.inner_coercers = [_as_instance(ic, CoercionToken) for ic in inner_coercers]

    def __call__(self, cls, filler):
        functors = [filler.get_coercer(t) for t in self.inner_coercers]
        functors.reverse()

        def ret(v):
            for f in functors:
                v = f(v)
            return v

        return ret



from typing import Optional, Type, TypeVar

T = TypeVar('T')


def _as_instance(obj, cls: Type[T]) -> Optional[T]:
    """
    Check that an object is of a type, or construct if it is a class.
    :param obj: The object to check and coerce.
    :param cls: The target class to check against.
    :return: `obj`, if it is an instance of `cls`, `obj()` if it is a subclass of `cls`, or `None` otherwise.
    """
    if isinstance(obj, cls):
        return obj
    elif isinstance(obj, type) and issubclass(obj, cls):
        return obj()
    wrapped = getattr(obj, '__wrapped__', None)
    if wrapped is not None:
        return obj()
    return None



import warnings
from abc import ABC, abstractmethod
from logging import Logger
from typing import Any, Callable, Generic, Type, TypeVar, Union


class ValidationToken:
    """
    A base class for all object that should be interpreted as validators
    """
    pass


T = TypeVar('T')


class GlobalValidationToken(ValidationToken, ABC):
    """
    A base class for all validators that can act on multiple kinds of fillers
    """

    @abstractmethod
    def __call__(self, stored_type: Type[T], filler) -> Callable[[Any], T]:
        """
        :param stored_type: the type the callback should return
        :param filler: the filler the validation callback will be called from
        :return: a callable to act as a validation callback for the specified filler
        """
        pass


class AssertValidation(GlobalValidationToken, ABC):
    """
    A validation token to check that a condition is upheld
    """

    def __init__(self, *, err: Union[str, Exception] = 'validation failed', warn: Union[bool, Logger] = False):
        """
        :param err: the error or warning to raise if the condition is not upheld.
        :param warn: whether to issue a warning instead if raising an error. Can also be an instance of a `Logger`,
         to specify the logger to warn with.
        """
        self.err = err
        self.warn = warn

    def _exc(self) -> Union[Exception, Warning]:
        """
        :return: The error or warning to raise when a condition is not upheld.
        """
        if isinstance(self.err, Exception):
            return self.err
        if self.warn:
            return UserWarning(self.err)
        return ValueError(self.err)

    def _raise(self):
        """
        raise the error or issue a warning
        """
        exc = self._exc()
        if self.warn:
            if self.warn is True:
                warnings.warn(exc)
            else:
                self.warn.warning(exc)
        else:
            raise exc

    @abstractmethod
    def assert_(self, v) -> bool:
        """
        :param v: the value to check.
        :return: Whether the condition is upheld for `v`.
        """
        pass

    def inner(self, v):
        """
        :param v: the value to validate
        """
        if not self.assert_(v):
            self._raise()
        return v

    def __call__(self, *_):
        return self.inner


class AssertCallValidation(AssertValidation, Generic[T]):
    """
    An assertion validation token to call arbitrary functions
    """
    def __init__(self, func: Callable[[T], bool], **kwargs):
        """
        :param func: the assertion function
        :param kwargs: forwarded to `AssertValidation.__init__`
        """
        super().__init__(**kwargs)
        self.func = func

    def assert_(self, v) -> bool:
        return self.func(v)


class CallValidation(GlobalValidationToken, Generic[T]):
    """
    An validation token to call arbitrary functions
    """
    def __init__(self, func: Callable[..., T], *args, **kwargs):
        """
        :param func: The callable to use as the validation callback.
        :param args: Optional positional arguments to pass to ``func``, after the validation argument.
        :param kwargs: Optional keyword arguments to pass to ``func``.
        .. note::
            calling ``CallValidation`` with ``args`` or ``kwargs`` is akin to calling it with a `functools.partial`_ as
             ``func``.
            >>> CallValidation(foo, a, b, c=d)
            >>> # is equivalent to
            >>> CallValidation(lambda v: foo(v, a, b, c=d))
        """
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def inner(self, v):
        return self.func(v, *self.args, **self.kwargs)

    def __call__(self, *_):
        return self.inner



from records.fillers.builtin_fillers.repo import builtin_filler_checkers, builtin_filler_map
from records.fillers.builtin_fillers.std_fillers import (Eval, LiteralEval, FromInteger, Loose, LooseUnpack,
                                                         LooseUnpackMap, Falsish, Whole,
                                                         std_filler_checkers, std_filler_map)
from records.fillers.builtin_fillers.typing_fillers import typing_checkers
from records.fillers.builtin_validators import Clamp, FullMatch, Truth, Within
from records.fillers.coercers import CallCoercion, ClassMethodCoercion, CoercionToken, ComposeCoercer, MapCoercion
from records.fillers.validators import AssertCallValidation, AssertValidation, CallValidation, ValidationToken

builtin_filler_map.update(std_filler_map)
builtin_filler_checkers.extend(std_filler_checkers)
builtin_filler_checkers.extend(typing_checkers)

__all__ = ['builtin_filler_map', 'builtin_filler_checkers',

           'CoercionToken', 'CallCoercion', 'MapCoercion', 'ClassMethodCoercion', 'ComposeCoercer',
           'ValidationToken', 'AssertValidation', 'CallValidation', 'AssertCallValidation',
           'Clamp', 'Within', 'FullMatch', 'Truth',

           'Eval', 'LiteralEval', 'Loose', 'LooseUnpack', 'LooseUnpackMap', 'Whole', 'FromInteger',
           'Falsish']



from typing import Hashable


class Tag:
    """
    A tag to mark a Field as belonging to a category. Tags encapsulate a single hashable object and implement only
     hashing and equality.
    """
    def __init__(self, x: Hashable):
        self.inner = x

    def __hash__(self):
        return hash(self.inner)

    def __eq__(self, other):
        return type(other) == type(self) \
               and self.inner == other.inner

    def __repr__(self):  # pragma: no cover
        return f'Tag({self.inner!r})'



from __future__ import annotations

import collections.abc as abstract_collections
import contextlib
from collections import defaultdict, deque
from collections.abc import Callable as CallableBase
from itertools import chain, islice
from typing import Any, Sequence, TypeVar, Union

from records.fillers.builtin_fillers.recurse import GetFiller
from records.fillers.filler import (AnnotatedFiller, Filler, TypeMatch, FillingSuccess, TypePassKind)
from records.fillers.get_filler import get_annotated_filler, get_filler
from records.utils.typing_compatible import get_args, get_origin

try:
    from typing import Literal
except ImportError:  # pragma: no cover
    Literal = object()


class UnionFiller(Filler):
    """
    A filler for union types.
    """

    def __init__(self, origin, args):
        super().__init__()
        self.args = args
        self.sub_types = get_args(origin)
        self.sub_fillers: Sequence[Filler] = ()
        self.applied = []

    def apply(self, token):
        super().apply(token)
        self.applied.append(token)

    def fill(self, arg):
        """
        The filling process for a Union is quite straight forward:
        * All of the union's inner arguments are parsed as fillers ion their own right, and are given the argument to
         type check
            * If any of these fillers succeed in exact type matching, only they are forwarded to the next stage.
            * Otherwise, if any of these fillers succeed in inexact type matching, only they are forwarded to the next
             stage.
            * Otherwise, if any of these fillers succeed in coercion, only they are forwarded to the next stage.
            * If no type checking succeed, the filling fails.
        * Then, all the successful sub-fillers are put through validation.
        * Validation succeeds only if exactly one sub-filler succeeded.
        """
        if self.is_hollow():
            return FillingSuccess(arg, TypePassKind.hollow)
        best_results = []
        best_key = float('inf')
        for i, sub_filler in enumerate(self.sub_fillers):
            try:
                success = sub_filler.fill(arg)
            except Exception:
                if not best_results and i + 1 == len(self.sub_fillers):
                    raise
                continue
            if success.type_pass_kind < best_key:
                best_results = [success.value]
                best_key = success.type_pass_kind
            elif success.type_pass_kind == best_key:
                best_results.append(success.value)

        if len(best_results) > 1:
            raise ValueError('multiple sub-fillers matched')
        return FillingSuccess(best_results[0], best_key)

    def bind(self, owner_cls):
        super().bind(owner_cls)
        self.sub_fillers = [get_filler(a) for a in self.sub_types]
        for sf in self.sub_fillers:
            for t in chain(self.args, self.applied):
                sf.apply(t)

            sf.bind(owner_cls)

    def is_hollow(self) -> bool:
        return all(sf.is_hollow() for sf in self.sub_fillers)


class LiteralFiller(AnnotatedFiller):
    """
    A filler for literal types.
    """

    def __init__(self, origin, args):
        super().__init__(origin, args)
        self.possible_values = defaultdict(set)
        for pv in get_args(origin):
            self.possible_values[type(pv)].add(pv)

    def type_check(self, v):
        if type(v) in self.possible_values:
            return TypeMatch.exact
        return isinstance(v, tuple(self.possible_values)) and TypeMatch.inexact

    def bind(self, owner_cls):
        super().bind(owner_cls)

        if not self.is_hollow():
            @self.validators.append
            def inner_validator(v):
                if v not in self.possible_values[type(v)]:
                    raise ValueError
                return v


class WrapperFiller(Filler):
    """
    A convenience class for fillers to wrap other fillers, with additional functionality (usually adding validators)
    """

    def __init__(self, origin, args):
        super().__init__()
        self.inner_filler = get_annotated_filler(origin, args)

    def fill(self, arg):
        return self.inner_filler.fill(arg)

    def bind(self, owner_cls):
        super().bind(owner_cls)
        self.inner_filler.bind(owner_cls)

    def is_hollow(self) -> bool:
        return self.inner_filler.is_hollow()

    def apply(self, token):
        self.inner_filler.apply(token)


class TypeFiller(WrapperFiller):
    """
    A filler for a parameterized typing.Type
    """

    def __init__(self, origin, args):
        super().__init__(get_origin(origin), args)
        self.base_type = get_args(origin)[0]

    def bind(self, owner_cls):
        super().bind(owner_cls)

        if not self.inner_filler.is_hollow():
            @self.inner_filler.validators.append
            def inner_validator(v):
                if not issubclass(v, self.base_type):
                    raise ValueError(f'must be a subclass of {self.base_type}')
                return v


class TupleFiller(WrapperFiller):
    """
    A filler for a parameterized typing.Tuple
    """

    def __init__(self, origin, args):
        super().__init__(get_origin(origin), args)
        self.inner_args = get_args(origin)

    def bind(self, owner_cls):
        super().bind(owner_cls)

        if len(self.inner_args) == 2 and self.inner_args[-1] is ...:
            inner_filler = get_filler(self.inner_args[0])
            inner_filler.bind(owner_cls)
            if not inner_filler.is_hollow():
                if self.is_hollow():
                    raise TypeError('cannot use non-hollow inner fillers in a hollow filler')

                @self.inner_filler.validators.append
                def inner_validator(v):
                    hollow_passes = 0  # the number of elements that filled identically
                    for element in v:
                        filled = inner_filler.__call__(element)
                        if filled is element:
                            hollow_passes += 1
                        else:
                            all_elements = chain(
                                v[:hollow_passes],
                                [filled],
                                (inner_filler.__call__(a) for a in v[hollow_passes + 1:])
                            )
                            return type(v)(all_elements)
                    return v
        else:
            if not self.is_hollow():
                @self.inner_filler.validators.append
                def _(v):
                    if len(v) != len(inner_fillers):
                        raise ValueError(f'must be a {len(inner_fillers)}-tuple')
                    return v

            inner_fillers = tuple(get_filler(t) for t in self.inner_args)
            for f in inner_fillers:
                f.bind(owner_cls)
            if not all(inner_filler.is_hollow() for inner_filler in inner_fillers):
                if self.is_hollow():
                    raise TypeError('cannot use non-hollow inner fillers in a hollow filler')

                @self.inner_filler.validators.append
                def _(v):
                    hollow_passes = 0  # the number of elements that filled identically
                    for element, filler in zip(v, inner_fillers):
                        filled = filler(element)
                        if filled is element:
                            hollow_passes += 1
                        else:
                            all_elements = chain(
                                v[:hollow_passes],
                                [filled],
                                (if_(a) for (a, if_) in zip(v[hollow_passes + 1:],
                                                            inner_fillers[hollow_passes + 1:]))
                            )
                            return type(v)(all_elements)
                    return v


def _split_at(seq, ignore_ind):
    """
    A utility function to split it in index, ignoring the element at the selected index
    :param seq: the sequence or iterable to split
    :param ignore_ind: the index to ignore
    :return: two iterables, one for `seq`'s elements up to `ignore_ind`, the second for all elements after.
    .. note::
        When used, the first iterable must be consumed entirely prior to the second one being consumed at all.
    """
    try:
        return seq[:ignore_ind], seq[ignore_ind + 1:]
    except TypeError:
        i = iter(seq)
        return islice(i, ignore_ind), islice(i, 1, None)


class GenericIterableFiller(WrapperFiller):
    """
    A filler for generic variants of concrete iterable classes.
    """

    def __init__(self, origin, args):
        super().__init__(get_origin(origin), args)
        self.inner_type = get_args(origin)[0]

    def reconstruct(self, elements, initial):
        """
        reconstruct a filled instance that passed type checking, after its elements have mutated
        :param elements: the new elements of the iterable
        :param initial: The original filled argument. Must not be mutated.
        :return: A new instance of `initial`'s type, with `elements` as its elements.
        """
        return type(initial)(elements)

    def bind(self, owner_cls):
        super().bind(owner_cls)

        inner_filler = get_filler(self.inner_type)
        inner_filler.bind(owner_cls)
        if not inner_filler.is_hollow():
            if self.is_hollow():
                raise TypeError('cannot use non-hollow inner fillers in a hollow filler')

            @self.inner_filler.validators.append
            def inner_validator(v):
                hollow_passes = 0  # the number of elements that filled identically
                for element in v:
                    filled = inner_filler(element)
                    if filled is element:
                        hollow_passes += 1
                    else:
                        pre, post = _split_at(v, hollow_passes)
                        all_elements = chain(
                            pre,
                            [filled],
                            (inner_filler(a) for a in post)
                        )
                        return self.reconstruct(all_elements, v)
                return v


class GenericDequeFiller(GenericIterableFiller):
    """
    A filler for generic deques
    """

    def reconstruct(self, elements, initial: deque):
        return type(initial)(elements, initial.maxlen)


class GenericMappingFiller(WrapperFiller):
    """
    A filler for generic variants of concrete mappings
    """

    def __init__(self, origin, args):
        super().__init__(get_origin(origin), args)
        self.key_type, self.value_type = get_args(origin)

    def reconstruct(self, tuples, initial):
        """
        reconstruct a filled instance that passed type checking, after its elements have mutated
        :param tuples: the new key-value tuples of the mapping
        :param initial: The original filled argument. Must not be mutated.
        :return: A new instance of `initial`'s type, with `tuples` as its tuples.
        """
        return type(initial)(tuples)

    def bind(self, owner_cls):
        super().bind(owner_cls)

        key_filler = get_filler(self.key_type)
        value_filler = get_filler(self.value_type)
        key_filler.bind(owner_cls)
        value_filler.bind(owner_cls)
        if not key_filler.is_hollow() or not value_filler.is_hollow():
            if self.is_hollow():
                raise TypeError('cannot use non-hollow inner fillers in a hollow filler')

            @self.inner_filler.validators.append
            def inner_validator(value):
                # there are no shortcuts here since maps can't be slices
                # (and thus this conversion would always require O(n) space)
                tuples = [
                    (key_filler(k), value_filler(v))
                    for k, v in value.items()
                ]
                if any(
                        (k is not filled_k) or (v is not filled_v)
                        for ((k, v), (filled_k, filled_v)) in zip(value.items(), tuples)
                ):
                    return self.reconstruct(tuples, value)
                else:
                    return value


class DefaultDictFiller(GenericMappingFiller):
    """
    A filler for generic defaultdicts
    """

    def reconstruct(self, tuples, initial: defaultdict):
        return type(initial)(initial.default_factory, tuples)


class GenericFillerN(WrapperFiller):
    """
    A filler for generic non-concrete classes or classes that otherwise can't have their parameters type-checked.
    """

    def __init__(self, origin, args):
        super().__init__(get_origin(origin), args)
        self.inner_types = get_args(origin)

    def bind(self, owner_cls):
        super().bind(owner_cls)

        fillers = [get_filler(it) for it in self.inner_types]
        for f in fillers:
            f.bind(owner_cls)
            if not f.is_hollow():
                raise TypeError('A non-specialized filler cannot have non-hollow inner types')


typing_checkers = []

genric_origin_map = {
    tuple: TupleFiller,
    type: TypeFiller,

    deque: GenericDequeFiller, defaultdict: DefaultDictFiller,

    abstract_collections.Sequence: GenericIterableFiller, abstract_collections.MutableSequence: GenericIterableFiller,
    abstract_collections.Set: GenericIterableFiller, abstract_collections.MutableSet: GenericIterableFiller,
    list: GenericIterableFiller, set: GenericIterableFiller, frozenset: GenericIterableFiller,

    abstract_collections.Iterable: GenericFillerN, abstract_collections.Iterator: GenericFillerN,
    abstract_collections.Reversible: GenericFillerN, abstract_collections.Collection: GenericFillerN,
    abstract_collections.Container: GenericFillerN, abstract_collections.AsyncIterator: GenericFillerN,
    abstract_collections.AsyncIterable: GenericFillerN, contextlib.AbstractContextManager: GenericFillerN,
    contextlib.AbstractAsyncContextManager: GenericFillerN,

    abstract_collections.Mapping: GenericMappingFiller, abstract_collections.MutableMapping: GenericMappingFiller,
    dict: GenericMappingFiller,

    abstract_collections.AsyncGenerator: GenericFillerN, abstract_collections.Generator: GenericFillerN
}


def has_args(v):
    args = get_args(v)
    return args and any(not isinstance(a, TypeVar) for a in args)


@typing_checkers.append
def _typing(stored_type):
    supertype = getattr(stored_type, '__supertype__', None)  # handle newtype
    if supertype:
        return GetFiller(supertype)

    if stored_type is Any:
        return GetFiller(object)

    origin_cls = get_origin(stored_type)
    if origin_cls == Union:
        return UnionFiller
    if origin_cls == Literal:
        return LiteralFiller
    if origin_cls == type:
        if has_args(stored_type):
            return TypeFiller
        return GetFiller(type)
    if origin_cls == CallableBase:
        return GetFiller(callable)
    if has_args(stored_type):
        t = genric_origin_map.get(origin_cls)
        if t:
            return t
    elif origin_cls:
        return GetFiller(origin_cls)



from typing import NamedTuple, Any


class GetFiller(NamedTuple):
    """
    An object to be returned by a builtin filler factory to specify that a filler should be retrieved from the
    specified type
    """
    new_origin: Any



from records.fillers.builtin_fillers.recurse import GetFiller
from records.fillers.builtin_fillers.repo import builtin_filler_checkers, builtin_filler_map
from records.fillers.filler import Filler, TypeCheckStyle, TypePassKind, FillingSuccess
from records.utils.typing_compatible import split_annotation


class DumbFiller(Filler):
    """
    A filler for objects that cannot be interpreted as types
    """

    def __init__(self, origin):
        super().__init__()
        self.origin = origin

    def fill(self, arg):
        return FillingSuccess(arg, TypePassKind.hollow)

    def bind(self, owner_cls):
        super().bind(owner_cls)
        if owner_cls.default_type_check_style() is not TypeCheckStyle.hollow:
            raise TypeError(f'cannot have type checking for origin type {self.origin}')

    def apply(self, token):
        raise TypeError(f'cannot have tokens for origin type {self.origin}')

    def is_hollow(self) -> bool:
        return True


def get_filler(stored_type) -> Filler:
    """
    get a filler for a type hint
    :param stored_type: the type annotation to use as the storage type
    :return: the `Filler` to fill targeting `storage_type`
    """
    origin, args = split_annotation(stored_type)
    return get_annotated_filler(origin, args)


def get_annotated_filler(origin, args: tuple):
    """
    get a filler for a type hint with annotations.
    :param origin: the origin storage type.
    :param args: Annotated arguments for the filler.
    :return:
    .. note::
        usually it's preferable to call `get_filler` with `Annotated`.
    """
    # there are 4 ways a filler is made:
    blt = builtin_filler_map.get(origin)
    if blt:
        # a mapping of types to filler classes (for shourtcuts)
        return blt(origin, args)
    for checker in builtin_filler_checkers:
        # builtin functions to create a filler class
        ret = checker(origin)
        if isinstance(ret, GetFiller):
            return get_annotated_filler(ret.new_origin, args)
        elif ret:
            return ret(origin, args)
    if isinstance(origin, type):
        # run through all classes in the mapping checking for subtypes
        blt_supertype = next((k for (k, v) in builtin_filler_map.items() if issubclass(origin, k)), None)
        if blt_supertype:
            return builtin_filler_map[blt_supertype](origin, args)
    if args:
        raise TypeError(f'cannot have Annotated type with origin {origin}')
    # finally, return a dumb filler
    return DumbFiller(origin)



from __future__ import annotations

from typing import ClassVar, Dict, Set, Union

from records.fillers import AssertCallValidation, CallCoercion, CallValidation
from records.fillers.filler import Filler
from records.fillers.get_filler import get_filler
from records.tags import Tag
from records.utils.decorators import decorator_kw_method
from records.utils.typing_compatible import get_args, get_origin, split_annotation

try:
    from typing import Final
except ImportError:
    Final = object()

NO_DEFAULT = object()
SKIP_FIELD = object()


class Factory:
    """
    A function wrapper to specify that a default value should be treated as a factory method
    """

    def __init__(self, func):
        self.func = func


class RecordField:
    """
    A singular field in a record, each field is owned by a single RecordBase subclass
    """

    def __init__(self, *, filler: Filler, owner, name: str, default):
        """
        :param filler: the filler instance to use when filling the field
        :param owner: the owning RecordBase subclass the field is bound to
        :param name: the name of the field
        :param default: the default of the field, wrapped in ``Factory`` for factory functions, or the singleton
             ``NO_DEFAULT`` if there is no default.
        """
        self.filler = filler
        self.name = name
        default, is_factory = self._apply_factory(default)

        self.default = default
        self.default_is_factory = is_factory
        self.owner: type = owner

        self.tags: Set[Tag] = set()

    def make_default(self):
        """
        get a default value of the field, either from the default or default factory

        :return: An instance default value for the field.

        .. warning:: It is an error to call this method on a field without a default
        """
        if self.default_is_factory:
            return self.default()
        return self.default

    @property
    def has_default(self):
        """
        :return: whether or not the field has a default set
        """
        return self.default is not NO_DEFAULT

    def is_default(self, v):
        """
        :param v: the value to compare
        :return: whether ``v`` is equal to the default value, if one exists
        .. note::
            currently, fields with a factory default always return ``False`` for this method, this is subject to change
        """
        return (not self.default_is_factory) and self.default == v

    @classmethod
    def _apply_factory(cls, default):
        if isinstance(default, Factory):
            return default.func, True
        return default, False

    @classmethod
    def from_type_hint(cls, th, *, owner, **kwargs) -> Union[RecordField, type(SKIP_FIELD)]:
        """
        Create a field from a type hint.
        :param th: the type hint to use
        :param owner: the owner RecordBase subclass
        :param kwargs: all keyword arguments are forwarded to the ``RecordField.__init__``
        :return: a RecordField instance matching the type hint provided, or the sentinel ``SKIP_FIELD`` to indicate
         that this declaration should be skipped (in case of a ClassVar)
        """
        if th is Final:
            raise TypeError('raw Final cannot be use inside records')
        origin, args = split_annotation(th)
        meta_org = get_origin(origin)
        if meta_org == ClassVar:
            return SKIP_FIELD
        if meta_org is Final:
            if not owner.is_frozen():
                raise TypeError('cannot declare Final field in non-frozen Record')
            ret = cls.from_type_hint(get_args(origin)[0], owner=owner, **kwargs)
        else:
            filler = get_filler(th)
            ret = cls(filler=filler, owner=owner, **kwargs)
        for arg in args:
            ret._apply(arg)
        return ret

    @decorator_kw_method
    def add_validator(self, func, **kwargs):
        """
        add a ``CallValidation`` to the field's filler.
        :param func: the callable to wrap inside the ``CallValidation``
        :param kwargs: all kwargs are forwarded to ``CallValidation``
        :return: func, to use as a decorator
        .. warning:: *must* be used inside of the owner class's `pre_bind` class method.
        .. code-block:: python

            class A(RecordBase):
                x: Annotated[int, check]
                @classmethod
                def pre_bind(cls):
                    super().pre_bind()
                    @cls.x.add_validator
                    def validator0():
                        ...

                    @cls.x.add_validator(a=0)
                    def validator1(a):
                        ...
        """
        self.filler.apply(CallValidation(func, **kwargs))
        return func

    @decorator_kw_method
    def add_assert_validator(self, func, **kwargs):
        """
        add a ``AssertCallValidation`` to the field's filler.
        :param func: the callable to wrap inside the ``AssertCallValidation``
        :param kwargs: all kwargs are forwarded to ``AssertCallValidation``
        :return: func, to use as a decorator
        .. warning:: *must* be used inside of the owner class's `pre_bind` class method.
        .. code-block:: python

            class A(RecordBase):
                x: Annotated[int, check]
                @classmethod
                def pre_bind(cls):
                    super().pre_bind()
                    @cls.x.add_assert_validator
                    def validator0():
                        ...

                    @cls.x.add_assert_validator(warn=True)
                    def validator1():
                        ...
        """
        self.filler.apply(AssertCallValidation(func, **kwargs))
        return func

    @decorator_kw_method
    def add_coercer(self, func, **kwargs):
        """
        add a ``CallCoercion`` to the field's filler.
        :param func: the callable to wrap inside the ``CallCoercion``
        :param kwargs: all kwargs are forwarded to ``CallCoercion``
        :return: func, to use as a decorator
        .. warning:: *must* be used inside of the owner class's `pre_bind` class method.
        .. code-block:: python

            class A(RecordBase):
                x: Annotated[int, check]
                @classmethod
                def pre_bind(cls):
                    super().pre_bind()
                    @cls.x.add_coercer
                    def coercion0():
                        ...

                    @cls.x.add_coercer(a=15)
                    def coercion1(a):
                        ...
        """
        self.filler.apply(CallCoercion(func, **kwargs))
        return func

    def _apply(self, token):
        if isinstance(token, Tag):
            self.tags.add(token)


class FieldDict(Dict[str, RecordField]):
    """
    A mapping from names to fields
    """

    def filter_by_tag(self, tag: Tag):
        """
        Filter the fields in the mapping to only those that have a tag
        :param tag: the tag to include
        :return: a new ``FieldDict`` including only the fields that possess tag_
        """
        return FieldDict((k, f) for (k, f) in self.items() if (tag in f.tags))



from __future__ import annotations

from abc import abstractmethod
from enum import Enum, IntEnum, auto
from typing import Any, Callable, Generic, List, Optional, TypeVar, NamedTuple

from records.fillers.coercers import CoercionToken
from records.fillers.util import _as_instance
from records.fillers.validators import ValidationToken

"""
Each field filling has multiple stages:
* Type checking (strict or non-strict)
    Makes sure the input is of the same type as the stored type. Strict checking only accepts values of the given type,
     and not sub-types. Strict type checking is illegal on abstract classes.
    Type checking will run only if:
        The type checking style has been set to Check, Check_strict, or Coerce
    The outcome of type checking:
        On success, report a no-coerce success to the parent
        On failure:
            * if the type checking style is Coerce, report a possible coerce match to the parent
            * if the type checking style does not allow coercion, report a failure to the parent
* coercion
    Attempts to coerce the value into the stored type. If the coercion function is run, it can be assumed that type
     checking has failed. The default coercers will perform the minimum amount of work possible. coercers SHOULD raise a
     TypeError if the value cannot be coerced. users can use custom coercers
    Coercion will run only if:
        The type checking style has been set to Coerce AND type checking has failed AND no other parallel filler has
        reported a no-coerce success
    The outcome of coercion:
        On success, report a coercion success
        On failure, report a failure
* validation
    After the type of the input has either been confirmed to be of the type, or coerced to the type, validators ensure
     the value matches a use-specific constraint, and change it or throw errors if it does not.
    A validation will only run if:
        a filler has not been aborted and has valdiators
    The outcome of validation:
        if a value is returned, a validation success is reported
        if an exception is raised, a validation error is reported
"""


class TypeCheckStyle(Enum):
    """
    Different ways to handle type-checking in fillers.
    """
    default = auto()
    """signifies that the filler should assume the type checking style of the owner record class upon binding.
     It is an error to have this style after binding"""
    hollow = auto()
    """signifies that no type checking or coercion should be performed"""
    check = auto()
    """signifies the filler should accept any argument of the type or its subclasses as valid arguments,
     and otherwise to fall back on coercion"""
    check_strict = auto()
    """signifies the filler should accept any argument of the exact type as valid arguments, and otherwise to fall
     back on coercion"""


class TypePassKind(IntEnum):
    """
    an inner enum to signify how a value passed a filler's type checking
    """
    no_coerce_strict = 0
    no_coerce = 1
    coerce = 2
    hollow = 3


T = TypeVar('T')


class FillingSuccess(NamedTuple):
    value: Any
    type_pass_kind: TypePassKind


class Filler(Generic[T]):
    """
    A base class for all fillers who must constrain arbitrary arguments to a (possibly annotated) storage type
    """

    def __init__(self):
        self.owner = None

    @abstractmethod
    def fill(self, arg) -> FillingSuccess[T]:
        """
        Begin filling a value to match the filler's type checking and validation
        :param arg: the initial value to use
        :return: a ``FillingSuccess`` representing a successful filling process
        """
        pass

    @abstractmethod
    def bind(self, owner_cls):
        """
        Bind a filler and associate it with a class. After binding, no new tokens can be applied to the filler.
        :param owner_cls: the owner record class to associate the filler with.
        """
        if self.owner:
            raise RuntimeError(f'filler is already bound to {self.owner}')
        self.owner = owner_cls

    @abstractmethod
    def is_hollow(self) -> bool:
        """
        :return: Whether or not the filler is "hollow". A hollow filler is one that accepts input of any type.
        """
        pass

    @abstractmethod
    def apply(self, token):
        """
        Apply a token to a filler.
        :param token: The token provided to the filler through `Annotated`.
        """
        if self.owner:
            raise RuntimeError('cannot apply tokens to filler after binding')

    def __call__(self, arg) -> T:
        """
        Coerce and validate an argument by the filler
        :param arg: The argument to fill with
        :return: `arg` after coercion and validation.
        """
        tc = self.fill(arg)
        return tc.value


class TypeMatch(Enum):
    """
    How a value matches a target storage type
    """
    exact = auto()
    """
    An exact type match
    """
    inexact = auto()
    """
    An inexact type match
    """


class AnnotatedFiller(Filler, Generic[T]):
    """
    A simple parent class for many filler subclasses
    """

    def __init__(self, origin, args):
        """
        :param origin: The origin of the `Annotated` storage type
        :param args: The arguments of the `Annotated` storage type
        .. note::
            If the storage type is not an `Annotated`, `origin` should be the storage type and `args` should be an
            empty tuple.
        """
        super().__init__()
        self.type_checking_style = TypeCheckStyle.default
        self.origin = origin
        self.args = args
        self.coercers: List[Callable[[Any], T]] = []
        self.validators: List[Callable[[T], T]] = []

    def fill(self, arg):
        if self.type_checking_style == TypeCheckStyle.default:  # pragma: no cover
            raise Exception
        elif self.type_checking_style == TypeCheckStyle.hollow:
            tpk = TypePassKind.hollow
        else:
            tp = self.type_check(arg)
            if tp == TypeMatch.exact:
                tpk = TypePassKind.no_coerce_strict
            elif tp == TypeMatch.inexact and self.type_checking_style == TypeCheckStyle.check:
                tpk = TypePassKind.no_coerce
            else:
                # perform coercion
                if not self.coercers:
                    raise TypeError(f'failed type checking for value of type {type(arg)}')
                for i, coercer in enumerate(self.coercers):
                    try:
                        arg = coercer(arg)
                        tc = self.type_check(arg)
                        if tc is None \
                                or (tc == TypeMatch.inexact and self.type_checking_style != TypeCheckStyle.check):
                            raise TypeError(f'coercer returned value of wrong type: {type(arg)}')
                    except Exception:
                        if i == len(self.coercers) - 1:
                            raise
                    else:
                        break
                tpk = TypePassKind.coerce

        # validation
        for validator in self.validators:
            arg = validator(arg)

        return FillingSuccess(arg, tpk)

    def bind(self, owner_cls):
        for arg in self.args:
            self.apply(arg)
        super().bind(owner_cls)
        if self.type_checking_style == TypeCheckStyle.default:
            self.type_checking_style = owner_cls.default_type_check_style()
        if self.type_checking_style == TypeCheckStyle.hollow and self.coercers:
            raise ValueError('cannot have hollow type checking with coercers')

    def apply(self, token):
        super().apply(token)
        if isinstance(token, TypeCheckStyle):
            self.type_checking_style = token
            return
        coercion_token = _as_instance(token, CoercionToken)
        if coercion_token is not None:
            self.coercers.append(self.get_coercer(coercion_token))
            return
        validation_token = _as_instance(token, ValidationToken)
        if validation_token is not None:
            self.validators.append(self.get_validator(validation_token))
            return

    @abstractmethod
    def type_check(self, v) -> Optional[TypeMatch]:
        """
        Check whether a value matches the storage type
        :param v: the value to check
        :return: The `TypeMatch` kind in case of a match, or `None` if there is no type match.
        """
        pass

    def get_coercer(self, token: CoercionToken) -> Callable[[Any], T]:
        """
        Get a coercion callback from a coercion token.
        :param token: the token to get the callback from.
        :return: A coercion callback originating from the token.
        :raises TypeError: If the token cannot be interpreted for this filler.
        """
        if callable(token):
            return token(self.origin, self)
        raise TypeError(token)  # pragma: no cover

    def get_validator(self, token: ValidationToken) -> Callable[[T], T]:
        """
        Get a validation callback from a validation token.
        :param token: the token to get the callback from.
        :return: A validation callback originating from the token.
        :raises TypeError: If the token cannot be interpreted for this filler.
        """
        if callable(token):
            return token(self.origin, self)
        raise TypeError(token)  # pragma: no cover

    def is_hollow(self) -> bool:
        return self.type_checking_style == TypeCheckStyle.hollow



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['records',
 'records.fillers',
 'records.fillers.builtin_fillers',
 'records.utils']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'recordclasses',
    'version': '0.0.1',
    'description': 'Record Classes',
    'long_description': '![Test records](https://github.com/bentheiii/Records/workflows/Test%20records/badge.svg?branch=master)\n[![codecov](https://codecov.io/gh/bentheiii/Records/branch/master/graph/badge.svg)](https://codecov.io/gh/bentheiii/Records)\n\n# Records\nRecords is a python library that makes powerful structure classes easy.\n\n## Simplest Example\nA feature of Records is that, by default, it does nothing more than a namedtuple or dataclass would.\n```python\nfrom records import RecordBase\n\nclass Point(RecordBase):\n    x: float\n    y: float\n    z: float = 0.0\n\np0 = Point(x=0, y=0)\nprint(p0)  # Point(x=0, y=0)\nprint(p0.x)  # 0\n# note that no type checking or coercion is performed\nprint(type(p0.y))  # int\n# by default, the type hints are not even run\np1 = Point(x="hello", y="world", z="1.0")\nprint(p1.y)  # world\nprint(type(p1.z))  # str\n```\n\n## Checking, Coercion, and Validation\nSometimes we\'d like to perform some additional processing to arguments before they are entered into the struct. For that we have three steps: type-checking, coercion, and validation\n* type-checking is the first and simplest step, it simply checks that the argument is of the type we expect. If it isn\'t, then we perform coercion.\n* type coercion only occurs if typechecking has failed, it will attempt to convert the into the type we expect. As one may expect, there is a large number of potential coercers, so they must be added each individually.\n* validation occurs after either type checking or type coercion have succeeded. By this point we are certain that the input is of the correct type and we want to ensure/manipulate its value.\n\nThese methods are described per-field with the `Annotation` type hint (`Annotation` is new in python 3.9, but has been backported by `records` for older versions to use)\n\nThese three steps are expanded upon in the documentation, for now we will show some brief examples:\n```python\nfrom typing import List\nfrom records import RecordBase, Annotated, check, check_strict, Loose, Within, Eval\n\nclass Person(RecordBase):\n    first_name: str  # no coercion or checking here, this is what we call a "hollow" field\n    last_name: Annotated[str, check]  # now we will raise a TypeError if anyone tries to enter a non-string last_name  \n    year_of_birth: Annotated[int, check_strict]  # we will raise a TypeError if year_of_birth isn\'t exactly an int (so passing True will throw an error)\n    lucky_number: Annotated[int, check, Loose]  # the "Loose" built-in coerser will simply call the destination type with the input as an argument, so that using `lucky_number="7"` would be equivalent to `lucky_number=int("7")`\n    number_of_children: Annotated[int, check, Within(ge=0)]  # the Within built-in validator ensures the value is within stated bounds (in this case, at least zero)\n    # field tokens can even be more complex in case of nested field types\n    names_of_children: Annotated[List[Annotated[int, check, Eval]], check]  # the list will be checked to be a list, and each item individually will be checked or coerced to be an int using the built-in Eval coercer.\n    \n    # validators can also be added after declaration with pre_bind\n    @classmethod\n    def pre_bind(cls):\n        @cls.last_name.add_validator\n        def no_bad_words(last_name):\n            # we want to remove some words from the last name\n            return last_name.replace(\'richard\', \'*******\')\n\n    # we can also add some more pre-processing on an entire instance with "post_new"\n    def post_new(self):\n        if len(self.names_of_children) != self.number_of_children:\n            raise ValueError("children mismatch")\n```\n## Parsing\nRecords can also be parsed from various python primitives. Including parsing from `dict`s, jsons, and even generic namepaces.\n```python\nfrom types import SimpleNamespace\nfrom records import RecordBase, check\n\nclass User(RecordBase, default_type_check=check):\n    name: str\n    password: str\n    age: int = 18\n    \n\nprint(User.from_mapping({"name": "richard", "password": "swordfish"}))\nprint(User.from_json(\'{"name": "richard", "password": "swordfish"})\'))\nn = SimpleNamespace(user="rich", password="ard", age= 7)\nprint(User.from_instance(n))\n\n# parsing can even be done if you expect misnamed fields!\nfrom_upper_dict = User.from_mapping.select(keys_to_rename=[(\'user\',\'name\')], keys_to_remove=[\'favorite_color\'])\nprint(from_upper_dict({\'user\':\'richard\', \'password\': \'pw\', \'favorite_color\': \'red\'}))\n```\n\nYou can also define your own parsers and even use them in construction!\n```python\nfrom math import sqrt\nfrom records import RecordBase, check, SelectableFactory, parser\n\nclass Point(RecordBase, default_type_check=check):\n    x: float\n    y: float\n    z: float = 0\n    \n    @parser\n    @SelectableConstructor\n    @classmethod\n    def from_tuple(cls, v):\n        return {\'x\':v[0], \'y\':v[1], \'z\':v[2] if len(v) > 2 else 0}\n\n    @property\n    def norm(self):\n        return sqrt(self.x**2 + self.y**2 + self.z**2)\n\np = Point([2,3,6])\nprint(p.norm)  # 7\n```\n## Exporting\nRecords can also be exporting to various formats (same as parsing).\n```python\nfrom records import RecordBase, check\n\nclass Point(RecordBase, default_type_check=check):\n    x: float\n    y: float\n    z: float\n\np = Point(x=2, y=3, z=6)\nprint(p.to_dict())\nprint(p.to_pickle())\n#  again, we can select to change the keys\nprint(\n    p.to_json.select(keys_to_add=[(\'w\',0)])()\n)\n```',
    'author': 'Ben Avrahami',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bentheiii/Records',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from collections import ChainMap
from copy import deepcopy
from inspect import getattr_static
from typing import AbstractSet, Any, Callable, ClassVar, Container, Dict, List, Mapping, Optional, TypeVar, Union, \
    NamedTuple
from warnings import warn

import records.extras as extras
from records.field import NO_DEFAULT, SKIP_FIELD, FieldDict, RecordField
from records.fillers.filler import TypeCheckStyle
from records.select import Exporter, NoArgExporter, SelectableFactory, SpecializedShortcutFactory, Select
from records.tags import Tag
from records.utils.typing_compatible import get_type_hints

try:
    from typing import Final
except ImportError:  # pragma: no cover
    Final = object()

NO_ARG = object()

exclude_from_ordering = Tag(object())


def parser(func: Callable):
    """
    mark a callable as a parser for a record class and all its subclasses
    :param func: the function to mark
    :return: ``func``, to used as a decorator
    """
    func.__parser__ = True
    return func


class ParseFailure(TypeError):
    """
    The error raised when an attempt to parse an object has failed because no parsers succeeded
    """
    pass


T = TypeVar('T')
FORBIDDEN_CLASS_ATTRS = ('__init__', '__setattr__', '__hash__')


# noinspection PyNestedDecorators
class RecordBase:
    __slots__ = '_hash',

    # a dict of fields, by name
    _fields: ClassVar[FieldDict]
    # a (immutable) set of field names that have no default value
    _required_keys: ClassVar[AbstractSet[str]]
    # a (immutable) set of field names that have a default value
    _optional_keys: ClassVar[AbstractSet[str]]
    # the default type check style that fillers can use if they have none defined
    _default_type_check_style: ClassVar[TypeCheckStyle]
    # whether the class is frozen
    _frozen: ClassVar[bool]
    # whether to allow unary parsing in constructor
    _unary_parse: ClassVar[bool]
    # a mutable list of parsers
    _parsers: ClassVar[List[Callable]]
    # whether the class is ordered
    _ordered: ClassVar[bool]

    def __init_subclass__(cls, *, frozen: bool = False, unary_parse: Optional[bool] = None, ordered=False,
                          default_type_check=TypeCheckStyle.hollow, **kwargs):
        """
        sets up the record subclass
        :param frozen: Whether the class should be considered immutable (and thus, hashable)
        :param unary_parse: Whether to enable unary parsing. By default is only disabled if the class has exactly one
         required field.
        :param default_type_check: The default type checking style of the class, all fields will use this style unless
         otherwise specified.
        """
        super().__init_subclass__(**kwargs)

        for attr in FORBIDDEN_CLASS_ATTRS:
            if attr in cls.__dict__:
                warn(f'in class {cls}: must not override {attr} (method will be overridden)')

        cls._frozen = frozen
        cls._ordered = ordered
        cls._default_type_check_style = default_type_check
        cls._fields = FieldDict()
        # in bodyless classes, __annotations__ refers to parent
        own_annotations = cls.__dict__.get('__annotations__', {})
        for name, type_hint in get_type_hints(cls, localns={cls.__name__: cls, cls.__qualname__: cls}).items():
            if name not in own_annotations:
                # type hints also give things from super classes which we don't want
                continue
            # initialize the field
            field = RecordField.from_type_hint(type_hint,
                                               owner=cls, name=name,
                                               default=getattr_static(cls, name, NO_DEFAULT))
            if field is SKIP_FIELD:
                continue
            cls._fields[name] = field
            setattr(cls, name, field)

        # initialize inherited fields separately
        parents = [b for b in cls.__bases__ if issubclass(b, RecordBase) and b != RecordBase]
        parent_fields = {}
        for parent in parents:
            field: RecordField
            for k, field in parent._fields.items():
                if k in cls._fields or k in parent_fields:
                    raise ValueError(f'cannot override inherited field {k}')
                parent_fields[k] = field
        if parent_fields:
            cls._fields = {**parent_fields, **cls._fields}

        if any(b for b in cls.__bases__ if b.__init__ not in (RecordBase.__init__, object.__init__)):
            warn(f'class {cls} has parents that implement __init__, the initializer will not be called!')

        if not cls._fields:
            raise ValueError(f'class {cls.__name__} has no fields')

        cls._required_keys = {k for (k, f) in cls._fields.items() if not f.has_default}
        cls._optional_keys = {k for k in cls._fields if k not in cls._required_keys}
        if unary_parse is None:
            # be default, unary parse is only enabled if that is the only way to treat a positional parameter
            unary_parse = len(cls._required_keys) != 1
        cls._unary_parse = unary_parse
        if frozen:
            cls.__setattr__ = RecordBase.__setattr__
            cls.__hash__ = RecordBase.__hash__
        else:
            cls.__setattr__ = object.__setattr__
            cls.__hash__ = None

        # all validator tokens go here
        cls.pre_bind()

        for field in cls._fields.values():
            if field.owner is not cls:
                continue
            field.filler.bind(cls)

        cls._parsers = []
        for name in dir(cls):
            v = getattr_static(cls, name)
            if getattr(v, '__parser__', False):
                cls._parsers.append(getattr(cls, name))

    @classmethod
    def pre_bind(cls):
        """
        A class method that gets called when the class is initialized, but before all the field fillers are bound to
         the class. Subclasses may override this method and add validators and coercers to fields.
        """
        pass

    def __new__(cls, arg=NO_ARG, **kwargs):
        """
        Create an instance of the class.
        :param arg: If there is exactly one required field in the class (called the trivial field), the single
         positional argument can be used to fill it. Alternatively, if ``unary_parse`` has been enabled, ``arg`` can be
         used (without keyword arguments) to parse the argument.
        :param kwargs: The mapping is used to fill the values of all fields in the record instance.
        .. note::
            If ``arg`` can be interpreted as both a parsing argument and as the trivial field. A ``TypeError`` is
             raised.
        """
        # parsing stores the parsing result, if any
        parsing = None
        if arg is not NO_ARG:
            # handle the positional
            if not kwargs and cls._unary_parse:
                try:
                    parsing = cls.parse(arg)
                except ParseFailure:
                    # if we failed parsing, and we have no trivial field, raise an error
                    if len(cls._required_keys) != 1:
                        raise

            if len(cls._required_keys) != 1:
                # no trivial field
                if parsing is not None:
                    return parsing
                raise TypeError(f'class {cls.__qualname__} accepts no positional arguments')

            # we have a trivial field
            arg_key = next(iter(cls._required_keys))
            if arg_key in kwargs:
                raise TypeError(f'duplicate {arg_key}')
            kwargs[arg_key] = arg

        values = {}
        # keep track of which field are missing
        required = set(cls._required_keys)
        optional = set(cls._optional_keys)
        try:
            for k, v in kwargs.items():
                required.discard(k)
                optional.discard(k)
                field = cls._fields.get(k)
                if not field:
                    raise TypeError(f'argument {k} invalid for type {cls.__qualname__}')
                values[k] = field.filler(v)
        except Exception:
            # if filling failed, check if we have a parsing standing by
            if parsing is not None:
                return parsing
            raise

        if required:
            raise TypeError(f'missing required arguments: {tuple(required)}')
        for k in optional:
            values[k] = cls._fields[k].make_default()

        self = super().__new__(cls)
        # we set directly into __dict__ because the class may be frozen and setattr would fail us
        self.__dict__.update(values)
        if cls._frozen:
            self._hash = None

        try:
            self = self.post_new() or self
        except Exception:
            # post_new may perform validations and if it fails, we want to return the parsing
            if parsing:
                return parsing
            raise

        # if we have successfully create an instance with positional argument, but still managed parsing, the we fail
        # we keep this check till last but
        if parsing is not None:
            raise TypeError(f'positional argument of type {type(arg)} can be interpreted both as trivial argument '
                            'and as unary argument')
        return self

    def post_new(self: T) -> Optional[T]:
        """
        This method is called after an instance is created and all its fields filled. This method may throw an exception
         to signal an invalid configuration.

        :return: May return a new instance, in which case it will replace the instance created, or None to
         keep it as is.
        """
        pass

    def __init__(self, arg=NO_ARG, **kwargs):
        """
        this init method ensures no other init methods run
        """
        pass

    def __repr__(self, **kwargs):
        params_parts = []
        for name, value in self.to_dict(**kwargs).items():
            params_parts.append(f'{name}={value!r}')

        return type(self).__qualname__ + "(" + ", ".join(params_parts) + ")"

    def __eq__(self, other):
        if type(self) != type(other):
            return False
        if type(self).is_frozen() and hash(self) != hash(other):
            return False
        for name in self._fields:
            if getattr(self, name) != getattr(other, name):
                return False
        return True

    class _MockField(NamedTuple):
        """
        A "virtual" field used when extracting data, to mimic an actual field
        """
        name: str

        tags = frozenset()
        has_default = False

    @classmethod
    def _to_dict(cls, obj, include_defaults=False, sort=None,
                 blacklist_tags: Union[Container[Tag], Tag] = frozenset(),
                 whitelist_keys: Union[Container[str], str] = frozenset(),
                 _rev_select: Select = Select.empty) -> Dict[str, Any]:
        """
        Create a dict representing the values of the class's fields for an arbitrary objects
        :param obj: the object to get attributes from
        :param include_defaults: whether to include keys that match the default value.
        :param sort: whether to sort the keys of the dictionary numerically. If falsish, the keys will not be sorted.
         If equal to -1, the items will be sorted in reverse lexicographic order. If callable, the items will be sorted
         according to ``sort`` as a key. Otherwise, the items will be sorted lexicographically.
        :param blacklist_tags: A ``Tag`` or set of ``Tag``s to ignore all fields with the ``Tag``.
        :param whitelist_keys: A field name or set of field names to include regardless of ``blacklist`` and
         ``include_defaults``.
        :param _rev_select: A private `Select`, intended to be applied over the result, the function will attempt to
         extract the appropriate keys to fit the select
        :return: A ``dict`` object with key names as specified by ``include_defaults``, ``sort``, ``blacklist_tags``,
         ``whitelist_keys``, and with values according to ``obj``'s attributes.
        :raises AttributeError: if ``obj`` lacks an attribute that has not been blacklisted
        """
        if isinstance(blacklist_tags, Tag):
            blacklist_tags = frozenset([blacklist_tags])
        if isinstance(whitelist_keys, str):
            whitelist_keys = frozenset([whitelist_keys])

        # to save as many attribute accesses as we can, we pre-filter any field we can filter without knowing
        # their value.

        def export_field(f: RecordField):
            # check if we can rule out a field without even knowing its value
            return f.name in whitelist_keys or not (f.tags & blacklist_tags)

        def export_value(f: RecordField, v):
            return f.name in whitelist_keys \
                   or include_defaults or not (f.has_default and f.is_default(v))

        missing_ok = cls._optional_keys
        extraction_dict = cls._fields

        if _rev_select:
            missing_ok = set(missing_ok)
            extraction_dict = dict(extraction_dict)
            for to_remove in _rev_select.keys_to_remove:
                if to_remove in cls._fields:
                    continue
                extraction_dict[to_remove] = cls._MockField(name=to_remove)
            for old, new in _rev_select.keys_to_rename:
                missing_ok.add(new)
                extraction_dict[old] = cls._MockField(name=old)
            for old, new in _rev_select.keys_to_maybe_rename:
                missing_ok.add(new)
                missing_ok.add(old)
                extraction_dict[old] = cls._MockField(name=old)
            for key, _ in _rev_select.keys_to_add:
                extraction_dict.pop(key, _)
        fields_to_extract = extraction_dict.values()

        def tuples():
            # inner function to lazily get the key-value tuples
            for field in fields_to_extract:
                if not export_field(field):
                    continue

                try:
                    v = getattr(obj, field.name)
                except AttributeError:
                    if field.name not in missing_ok:
                        # we allow skipping missing fields if they have a default
                        raise
                    if include_defaults and field.has_default:
                        # but if the field is missing, and we expect to find it anyway, then we yield its tuple even
                        # though it was not present
                        yield field.name, field.make_default()
                else:
                    if export_value(field, v):
                        yield field.name, v

        name_values = tuples()
        if sort:
            if sort == -1:
                name_values = sorted(name_values, reverse=True)
            elif callable(sort):
                name_values = sorted(name_values, key=lambda t: sort(t[0]))
            else:
                name_values = sorted(name_values)

        return dict(name_values)

    @classmethod
    def is_frozen(cls):
        """
        :return: Whether the class is frozen.
        """
        return cls._frozen

    @classmethod
    def default_type_check_style(cls):
        """
        :return: The default type checking style.
        """
        return cls._default_type_check_style

    def __hash__(self):
        # note: if the class is non-frozen, this function will be overridden
        if self._hash is None:
            self._hash = hash(tuple(self.to_dict().values()))
        return self._hash

    def __setattr__(self, a, value):
        # note: if the class is non-frozen, this function will be overridden
        if a in RecordBase.__slots__:
            return super().__setattr__(a, value)
        raise TypeError(f'{type(self).__qualname__} is frozen')

    def __getnewargs_ex__(self):
        return (), self.to_dict()

    @parser
    @SelectableFactory
    @classmethod
    def from_mapping(cls, *maps: Mapping[str, Any], **kwargs: Any):
        """
        Convert a mapping to a Record instance.
        :param maps: Mappings to combine into field values.
        :param kwargs: Additional field name in the instance.
        :return: An instance of ``cls`` with arguments as described by the input mappings.
        .. note::
            This class method supports `selection`_.
        .. note::
            This class method is a registered parser that will be attempted when calling ``cls.parse``.
        """
        return ChainMap(*maps, dict(**kwargs))

    @SpecializedShortcutFactory
    @classmethod
    def from_instance(cls, v, *maps: Mapping[str, Any], _select=Select.empty, **kwargs):
        """
        Convert an object to a Record instance by attributes.
        :param v: An object to get attributes from.
        :param maps: Mappings to combine into field values.
        :param _select: a private `Select` to be used when extracting object attributes.
        :param kwargs: Additional field name in the instance.
        :return: An instance of ``cls`` with arguments as described by the input namespace and mappings.
        .. note::
            This class method supports `selection`_.
        .. note::
            If the class is frozen, and there are no additional mappings or kwargs supplied, the method may
            return ``v``.
        .. note::
            This class method is a registered parser that will be attempted when calling ``cls.parse``.
        """
        d = cls._to_dict(v, _rev_select=_select)
        d.update(*maps, **kwargs)
        return d

    @parser
    @from_instance.shortcut
    @classmethod
    def from_instance(cls, v, *maps: Mapping[str, Any], **kwargs):
        if cls.is_frozen() and not maps and not kwargs and isinstance(v, cls):
            return v
        return NotImplemented

    @parser
    @SelectableFactory
    @classmethod
    def from_json(cls, v, **kwargs):
        """
        Convert a JSON mapping to a Record instance.
        :param v: An JSON dictionary string to get attributes from.
        :param kwargs: Additional field name in the instance.
        :return: An instance of ``cls`` with arguments as described by the JSON.
        .. note::
            This class method supports `selection`_.
        .. note::
            This class method is a registered parser that will be attempted when calling ``cls.parse``.
        """
        return extras.json.loads(v, **kwargs)

    @SelectableFactory
    @classmethod
    def from_json_io(cls, v, **kwargs):
        """
        Convert a JSON file to a Record instance.
        :param v: An JSON dictionary file string to get attributes from.
        :param kwargs: Additional field name in the instance.
        :return: An instance of ``cls`` with arguments as described by the JSON.
        .. note::
            This class method supports `selection`_.
        """
        return extras.json.load(v, **kwargs)

    @parser
    @classmethod
    def from_pickle(cls, v, *args, **kwargs):
        """
        Convert a pickled bytestring to a Record instance.
        :param v: A bytestring object.
        :param args: forwarded to `pickle.loads`_
        :param kwargs: forwarded to `pickle.loads`_
        :return: An unpickled instance of ``cls``.
        .. note::
            If the unpickling result succeeds but the result is nto an instance of ``cls``, then ``cls`` attempts
            to parse the resulting object.
        .. note::
            This class method is a registered parser that will be attempted when calling ``cls.parse``.
        """
        ret = extras.pickle.loads(v, *args, **kwargs)
        if not isinstance(ret, cls):
            return cls.parse(ret)
        return ret

    @classmethod
    def from_pickle_io(cls, v, *args, **kwargs):
        """
        Convert a pickled file to a Record instance.
        :param v: A record bytestring object.
        :param args: forwarded to `pickle.load`_
        :param kwargs: forwarded to `pickle.load`_
        :return: An unpickled instance of ``cls``.
        .. note::
            If the unpickling result succeeds but the result is nto an instance of ``cls``, then ``cls`` attempts
            to parse the resulting object.
        """
        ret = extras.pickle.load(v, *args, **kwargs)
        if not isinstance(ret, cls):
            return cls.parse(ret)
        return ret

    @classmethod
    def parse(cls, v):
        """
        Attempt to run all registered parsers of ``cls`` to parse an object to a ``cls`` instance.
        :param v: the object to attempt to parse.
        :return: The result of the only parser to succeed (raises an exception in all other cases).
        :raise ParseFailure: If none of the registered parsers succeed.
        :raise TypeError: If more than one of the registered parsers succeed.
        """
        successes = []
        for m in cls._parsers:
            try:
                r = m(v)
            except Exception:
                pass
            else:
                successes.append(r)
        if len(successes) == 0:
            raise ParseFailure(f'cannot parse {cls.__qualname__} from argument of type {type(v).__qualname__}')
        elif len(successes) > 1:
            raise TypeError(f'multiple unary constructors succeeded with argument of type {type(v).__qualname__}')

        return next(iter(successes))

    @NoArgExporter
    @staticmethod
    def to_dict(v) -> Mapping[str, Any]:
        """
        export an instance to a dictionary.
        .. note::
            This class method supports `selection`_ and `exporting arguments`_.
        """
        return v

    @Exporter
    @staticmethod
    def to_json(v, *args, io=None, **kwargs) -> str:
        """
        export an instance to a JSON dictionary.
        :param args: forwarded to either `json.dump`_ or `json.dumps`_
        :param io: If not ``None``, dumps ``self`` into ``io``.
        :param kwargs: forwarded to either `json.dump`_ or `json.dumps`_
        :return: ``None`` if ``io`` is not ``None``, otherwise a JSON string representing ``self``
        .. note::
            This class method supports `selection`_ and `exporting arguments`_.
        """
        if io is None:
            return extras.json.dumps(v, *args, **kwargs)
        extras.json.dump(v, io, *args, **kwargs)

    def to_pickle(self, *args, io=None, **kwargs) -> bytes:
        """
        export an instance to a pickled bytestring.
        :param args: forwarded to either `pickle.dump`_ or `pickle.dumps`_
        :param io: If not ``None``, dumps ``self`` into ``io``.
        :param kwargs: forwarded to either `pickle.dump`_ or `pickle.dumps`_
        :return: ``None`` if ``io`` is not ``None``, otherwise a pickle bytestring representing ``self``
        """
        if io is None:
            return extras.pickle.dumps(self, *args, **kwargs)
        extras.pickle.dump(self, io, *args, **kwargs)

    def __copy__(self):
        if self.is_frozen():
            return self
        return type(self).from_instance(self)

    def __deepcopy__(self, memo=None):
        memo = memo or {}
        d = self.to_dict()
        d = deepcopy(d, memo)
        return self.from_mapping(d)

    def __ordering_key(self):
        f: RecordField
        ret = []
        for k, f in type(self)._fields.items():
            if exclude_from_ordering in f.tags:
                continue
            ret.append(getattr(self, k))
        return ret

    def check_comparable(self, other):
        if not type(self)._ordered:
            raise TypeError(f'cannot compare with non-ordered type {type(self).__qualname__}')

    def __lt__(self, other):
        self.check_comparable(other)
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.__ordering_key() < other.__ordering_key()

    def __le__(self, other):
        self.check_comparable(other)
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.__ordering_key() <= other.__ordering_key()

    def __gt__(self, other):
        self.check_comparable(other)
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.__ordering_key() > other.__ordering_key()

    def __ge__(self, other):
        self.check_comparable(other)
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.__ordering_key() >= other.__ordering_key()



from functools import wraps


def decorator_kw_method(func):
    """
    A decorator to apply to other decorator methods, to allow them to delay calling with keyword arguments.
    .. example::
        >>> @decorator_kw_method
        >>> def foo(func, *, a=0):
        >>>     ...
        >>> @foo
        >>> def bar0():
        >>>     ...
        >>> @foo(a=3)
        >>> def bar3():
        >>>     ...
    """
    @wraps(func)
    def ret(self, f=None, **kwargs):
        if f is None:
            return lambda f: ret(self, f, **kwargs)
        return func(self, f, **kwargs)

    return ret



from __future__ import annotations

from functools import update_wrapper
from itertools import chain
from typing import Any, Callable, Generic, Iterable, Mapping, Optional, Tuple, Type, TypeVar, Union


class Select:
    """
    A Select is a method of modifying a ``Mapping[str, Any]``. A Select should be considered immutable
    """
    empty: Select
    """An Empty Select, to be used as a default value to represent no mutations over a mapping"""

    def __init__(self, *,
                 keys_to_add: Union[Iterable[Tuple[str, Any]], Mapping[str, Any]] = (),
                 keys_to_maybe_add: Union[Iterable[Tuple[str, Any]], Mapping[str, Any]] = (),
                 keys_to_remove: Union[Iterable[str], str] = (),
                 keys_to_maybe_remove: Union[Iterable[str], str] = (),
                 keys_to_rename: Union[Iterable[Tuple[str, str]], Mapping[str, str]] = (),
                 keys_to_maybe_rename: Union[Iterable[Tuple[str, str]], Mapping[str, str]] = (),
                 ):
        """
        :param keys_to_add: Keys to add to the mapping, raising an error if they already exist
        :param keys_to_maybe_add: Keys to add to the mapping, or skip if they already exist
        :param keys_to_remove: Keys to remove from the mapping, raising an error if they don't exist. If a single
         string is presented, it is interpreted as a single key to remove.
        :param keys_to_maybe_remove: Keys to remove from the mapping if they exist. If a single string is presented,
         it is interpreted as a single key to remove.
        :param keys_to_rename: Keys to whose value to transfer to other keys inside the mapping, raising an error if the
         new name already exists
        :param keys_to_maybe_rename: Keys to whose value to transfer to other keys inside the mapping, or skip if the
         new name already exists
        """
        if isinstance(keys_to_add, Mapping):
            keys_to_add = tuple(keys_to_add.items())
        if isinstance(keys_to_maybe_add, Mapping):
            keys_to_maybe_add = tuple(keys_to_maybe_add.items())
        if isinstance(keys_to_rename, Mapping):
            keys_to_rename = tuple(keys_to_rename.items())
        if isinstance(keys_to_maybe_rename, Mapping):
            keys_to_maybe_rename = tuple(keys_to_maybe_rename.items())
        if isinstance(keys_to_remove, str):
            keys_to_remove = (keys_to_remove,)
        if isinstance(keys_to_maybe_remove, str):
            keys_to_maybe_remove = (keys_to_maybe_remove,)
        self.keys_to_add = keys_to_add
        self.keys_to_maybe_add = keys_to_maybe_add
        self.keys_to_remove = keys_to_remove
        self.keys_to_maybe_remove = keys_to_maybe_remove
        self.keys_to_rename = keys_to_rename
        self.keys_to_maybe_rename = keys_to_maybe_rename

        # the Select's truthiness is cached
        self._bool = None

    def _merge(self, other: Select):
        if not self:
            return other
        if not other:
            return self
        return type(self)(
            keys_to_add=tuple(chain(self.keys_to_add, other.keys_to_add)),
            keys_to_maybe_add=tuple(chain(self.keys_to_maybe_add, other.keys_to_maybe_add)),
            keys_to_remove=tuple(chain(self.keys_to_remove, other.keys_to_remove)),
            keys_to_maybe_remove=tuple(chain(self.keys_to_maybe_remove, other.keys_to_maybe_remove)),
            keys_to_rename=tuple(chain(self.keys_to_rename, other.keys_to_rename)),
            keys_to_maybe_rename=tuple(chain(self.keys_to_maybe_rename, other.keys_to_maybe_rename)),
        )

    def merge(self, *others: Select, **kwargs):
        """
        combine several Selects into one
        :param others: other selects to combine with ``self``
        :param kwargs: additional arguments to create a select with, and merge it.
        :return: ``self``, all of ``others``, and a ``Select`` formed with ``kwargs``, combined into a single ``Select``
        """
        if not others and not kwargs:  # pragma: no cover
            raise TypeError('merge must be called with arguments')
        ret = self
        for other in others:
            ret = ret._merge(other)
        if kwargs:
            ret = ret._merge(type(self)(**kwargs))
        return ret

    def __bool__(self):
        if self._bool is None:
            self._bool = next(chain(
                self.keys_to_add,
                self.keys_to_maybe_add,
                self.keys_to_remove,
                self.keys_to_maybe_remove,
                self.keys_to_rename,
                self.keys_to_maybe_rename,
            ), None) is not None
        return self._bool

    def __call__(self, mapping: Mapping[str, Any]) -> Mapping[str, Any]:
        """
        :param mapping: The mapping to use as input
        :return: A modified ``Mapping`` as specified by ``self``
        .. warning::
            Any Mapping sent to this function should not be used anywhere else, as this function may modify it.
        """
        # this function may well alter the source mapping
        if not self:
            return mapping
        if not isinstance(mapping, dict):
            mapping = dict(mapping)

        for kti in self.keys_to_remove:
            mapping.pop(kti)
        for kti in self.keys_to_maybe_remove:
            mapping.pop(kti, None)
        for s, d in self.keys_to_rename:
            if d in mapping:
                raise ValueError(f'key {d} cannot be overridden in map')
            mapping[d] = mapping.pop(s)
        for s, d in self.keys_to_maybe_rename:
            if s not in mapping:
                continue
            if d in mapping:
                raise ValueError(f'key {d} cannot be overridden in map')
            mapping[d] = mapping.pop(s)
        for s, v in self.keys_to_add:
            if s in mapping:
                raise ValueError(f'key {s} cannot be overridden in map')
            mapping[s] = v
        for s, v in self.keys_to_maybe_add:
            if s in mapping:
                continue
            mapping[s] = v

        return mapping


Select.empty = Select()

T = TypeVar('T')


class SelectableFactory(Generic[T]):
    """
    A class to hold class factories that can be configured with the ``select`` method
    .. Note::
        Users can create their own SelectableFactory by using this class as decorator (on top of ``@classmethod``)
    """

    def __init__(self, func):
        """
        :param func: The function to wrap around, must return a mapping.
        """
        if isinstance(func, classmethod):
            func = func.__func__
        self.func: Callable[..., Mapping[str, Any]] = func
        update_wrapper(self, func)

    def run(self, cls: Type, args: Iterable, kwargs: Mapping[str, Any], select: Select):
        """
        run the factory with given arguments and selector
        :param cls: the owner class to create
        :param args: the arguments forwarded to ``func``
        :param kwargs: the keyword arguments forwarded to ``self.func``
        :param select: the select to apply on the mapping ``self.func`` returns
        :return: An instance of ``cls`` with arguments from ``self.func`` after ``select``.
        """
        mapping = self.func(cls, *args, **kwargs)
        mapping = select(mapping)
        return cls(**mapping)

    def __get__(self, instance, owner):
        """
        :return: The factory bound to an owner class
        """
        return self.Bound(self, owner, Select.empty)

    class Bound:
        """
        A bound instance of a ``SelectableFactory``
        """

        def __init__(self, descriptor: SelectableFactory, owner_cls: Type, select: Select):
            """
            :param descriptor: the ``SelectableFactory`` parent
            :param owner_cls: the owner class the instance is bound to
            :param select: the ``Select`` the instance should use
            """
            self.descriptor = descriptor
            self.owner_cls = owner_cls
            self.select_ = select

        def __call__(self, *args, **kwargs):
            return self.descriptor.run(self.owner_cls, args, kwargs, self.select_)

        def select(self, *selects: Select, **kwargs):
            """
            create a new bound factory with a modified ``Select``
            :param selects: the new `Select`_s to merge into the existing select
            :param kwargs: keyword argument to merge into the new `Select`_
            :return: a new bound factory
            """
            return type(self)(self.descriptor, self.owner_cls, self.select_.merge(*selects, **kwargs))


class SpecializedSelectableFactory(SelectableFactory):
    def run(self, cls: Type, args: Iterable, kwargs: Mapping[str, Any], select: Select):
        mapping = self.func(cls, *args, _select=select, **kwargs)
        mapping = select(mapping)
        return cls(**mapping)


class SelectableShortcutFactory(SelectableFactory):
    """
    A selectable factory with a shortcut function, that will be called if the ``select`` is false.
    """

    def __init__(self, func, shortcut: Optional[Callable] = None):
        super().__init__(func)
        if isinstance(shortcut, classmethod):
            shortcut = shortcut.__func__
        self._shortcut = shortcut

    def shortcut(self, sc):
        """
        Set a shortcut
        :param sc: the Shortcut function or class method. The shortcut function should either return an instance of the
         owner class or ``NotImplemented`` to indicate to fall back to main implementation.
        :return: a new ``SelectableShortcutFactory`` with ``sc`` as its shortcut.
        .. note::
            this function can be used as a decorator, provided that the shortcut function and main function have the
            same name.
        """
        if self._shortcut:  # pragma: no cover
            raise ValueError('cannot set multiple shortcuts')
        return type(self)(self.func, sc)

    def run(self, cls, args, kwargs, select: Select):
        if not select and self._shortcut:
            ret = self._shortcut(cls, *args, **kwargs)
            if ret is not NotImplemented:
                return ret
        return super().run(cls, args, kwargs, select)


class SpecializedShortcutFactory(SelectableShortcutFactory, SpecializedSelectableFactory):
    pass


class Exporter(Generic[T]):
    """
    An exported function that supports selection and additional exporting configuration
    """

    def __init__(self, func: Callable):
        """
        :param func: the wrapped function, should accept the first positional argument a mapping.
        """
        if isinstance(func, staticmethod):
            func = func.__func__
        self.func = func
        self.owner: Optional[type] = None

    def run(self, cls, instance, export_args, export_kwargs, select, args, kwargs):
        """
        runs the exporter
        :param instance: The instance to export
        :param export_args: the arguments of the exporting function `RecordBase._to_dict`_.
        :param export_kwargs: the keyword arguments of the exporting function `RecordBase._to_dict`_.
        :param select: the select to apply over the mapping before passing it to ``func``.
        :param args: the arguments of ``self.func``
        :param kwargs: the keyword arguments of ``self.func``
        :return: the result of ``func`` over the final mapping
        """
        mapping = cls._to_dict(instance, *export_args, **export_kwargs)
        mapping = select(mapping)
        return self.func(mapping, *args, **kwargs)

    def __get__(self, instance, owner):
        """
        Get a bound instance of an exporter function.
        """
        if instance is None:  # pragma: no cover
            return self.BoundToClass(self, owner, (), {}, Select.empty)
        return self.Bound(self, instance, (), {}, Select.empty)

    class Bound:
        """
        An exporter bound to a specific instance
        """

        def __init__(self, descriptor: Exporter, owner, export_args, export_kwargs, select):
            """
            :param descriptor: The exporter
            :param owner: the owner instance that the object is bound to
            :param export_args: the arguments passed to `RecordBase._to_dict`_
            :param export_kwargs: the keyword arguments passed to `RecordBase._to_dict`_
            :param select: the select to apply
            """
            self.descriptor = descriptor
            self.owner = owner
            self.export_args = export_args
            self.export_kwargs = export_kwargs
            self.select_ = select

        def __call__(self, *args, **kwargs):
            return self.descriptor.run(self.owner, self.owner, self.export_args, self.export_kwargs, self.select_, args,
                                       kwargs)

        def select(self, *selects: Select, **kwargs):
            """
            create a new bound exporter with a modified ``Select``
            :param selects: the new `Select`_s to merge into the existing select
            :param kwargs: keyword argument to merge into the new `Select`_
            :return: a new bound exporter
            """
            return type(self)(self.descriptor, self.owner, self.export_args, self.export_kwargs,
                              self.select_.merge(*selects, **kwargs))

        def export_with(self, *args, **kwargs):
            """
            create a new bound exporter with modified export arguments
            :param args: forwarded to `RecordBase._to_dict`_
            :param kwargs: forwarded to `RecordBase._to_dict`_
            :return: a new bound exporter
            """
            return type(self)(self.descriptor, self.owner, (*self.export_args, *args), {**self.export_kwargs, **kwargs},
                              self.select_)

    class BoundToClass(Bound):
        def __call__(self, instance, *args, **kwargs):
            return self.descriptor.run(self.owner, instance, self.export_args, self.export_kwargs, self.select_, args,
                                       kwargs)


class NoArgExporter(Exporter):
    """
    An exporter with the difference that ``func`` takes no arguments so all arguments are passed into
     `RecordBase._to_dict`_
    """

    def run(self, cls, instance, export_args, export_kwargs, select, args, kwargs):
        return super().run(cls, instance, (*export_args, *args), {**export_kwargs, **kwargs}, select, (), {})



"""
This file provides multiple packports of typing names that may not exist in other versions.
"""

import sys
from functools import partial
from typing import Callable, get_type_hints

if sys.version_info >= (3, 8, 0):  # pragma: no cover
    from typing import get_args, get_origin
else:  # pragma: no cover
    def get_origin(v):
        return getattr(v, '__origin__', None)

    def get_args(v):
        return getattr(v, '__args__', None)  # todo callable needs special case

if sys.version_info >= (3, 9, 0):  # pragma: no cover
    from typing import Annotated, _AnnotatedAlias

    get_type_hints = partial(get_type_hints, include_extras=True)

    def is_annotation(t):
        return isinstance(t, _AnnotatedAlias)
else:  # pragma: no cover
    class Annotated:
        def __init__(self, *args):
            self.__origin__ = type(self)
            self.__args__ = args

        def __class_getitem__(cls, item):
            return cls(*item)

        def __call__(self, *args, **kwargs):
            raise TypeError

    Annotated.__origin__ = Annotated

    def is_annotation(t):
        return isinstance(t, Annotated)

    if sys.version_info >= (3, 8, 0):  # pragma: no cover
        _origins = get_origin
        _args = get_args

        def get_origin(v):
            return _origins(v) or getattr(v, '__origin__', None)

        def get_args(v):
            if v is Callable:
                return ()
            return _args(v) or getattr(v, '__args__', None)


def split_annotation(v):
    if not is_annotation(v):
        return v, ()
    t, *args = get_args(v)
    return t, args


__all__ = ['get_args', 'get_origin', 'Annotated', 'get_type_hints', 'is_annotation', 'split_annotation']



"""
This module includes libraries to use that can be overridden by user programs. For instance, to make Records encode and
 decode with `ujson <https://github.com/ultrajson/ultrajson>`_ use the following code::

    import records.extras
    import ujson
    records.extras.json = ujson
"""

import json
import pickle
import re

__all__ = ['json', 'pickle', 're']
